# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2008 Gautier Hayoun <gautier.hayoun@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime
from tempfile import mkdtemp
from subprocess import call
from urllib import urlencode
from re import compile

# Import from docutils
from docutils.core import Publisher, publish_from_doctree, publish_string
from docutils.io import StringOutput, DocTreeInput
from docutils.readers.doctree import Reader
from docutils.utils import SystemMessage
from docutils import nodes

# Import from itools
from itools import vfs
from itools.datatypes import DateTime, String, Unicode
from itools.gettext import MSG
from itools.handlers import checkid, get_handler, File as FileHandler
from itools.html import XHTMLFile
from itools.i18n import format_datetime
from itools.uri import get_reference
from itools.uri.mailto import Mailto
from itools.utils import guess_extension
from itools.vfs import FileName
from itools.web import BaseView, STLForm, STLView, ERROR
from itools.xml import XMLParser, XMLError

# Import from ikaaro
from ikaaro import messages
from ikaaro.html import is_edit_conflict


figure_style_converter = compile(r'\\begin\{figure\}\[.*?\]')

class WikiPage_View(BaseView):

    access = 'is_allowed_to_view'
    title = MSG(u'View')
    icon = 'html.png'


    def GET(self, resource, context):
        context.styles.append('/ui/wiki/style.css')
        parent = resource.parent

        try:
            document = resource.get_document()
        except SystemMessage, e:
            # Critical
            context.message = ERROR(u'Syntax error: $error', error=e.message)
            return XMLParser('<pre>' + resource.handler.to_str() + '</pre>')

        # Decorate the links and resolve them against the published resource
        for node in document.traverse(condition=nodes.reference):
            refname = node.get('wiki_name')
            if refname is None:
                # Regular link
                if node.get('refid'):
                    node['classes'].append('internal')
                refuri = node.get('refuri')
                if refuri is not None:
                    reference = get_reference(refuri.encode('utf_8'))
                    # Skip external
                    if reference.scheme or reference.authority:
                        node['classes'].append('external')
                        continue
                    try:
                        destination = resource.get_resource(reference.path)
                        node['refuri'] = context.get_link(destination)
                    except LookupError:
                        pass
            elif refname is False:
                    # Wiki link not found
                    node['classes'].append('nowiki')
                    prefix = resource.get_pathto(parent)
                    title = node['name']
                    title_encoded = title.encode('utf_8')
                    params = {'type': resource.__class__.__name__,
                              'title': title_encoded,
                              'name': checkid(title) or title_encoded}
                    refuri = "%s/;new_resource?%s" % (prefix,
                                                      urlencode(params))
                    node['refuri'] = refuri
            else:
                # Wiki link found, "refname" is the path
                node['classes'].append('wiki')
                destination = resource.get_resource(refname)
                node['refuri'] = context.get_link(destination)

        # Manipulate publisher directly (from publish_from_doctree)
        reader = Reader(parser_name='null')
        pub = Publisher(reader, None, None, source=DocTreeInput(document),
                destination_class=StringOutput)
        pub.set_writer('html')
        pub.process_programmatic_settings(None, resource.overrides, None)
        pub.set_destination(None, None)
        pub.publish(enable_exit_status=None)
        parts = pub.writer.parts
        body = parts['html_body']

        return body.encode('utf_8')



class WikiPage_ToPDF(BaseView):

    access = 'is_allowed_to_view'
    title = MSG(u"To PDF")


    def GET(self, resource, context):
        parent = resource.parent
        pages = [resource.get_abspath()]
        images = []

        document = resource.get_document()
        # We hack a bit the document tree to enhance the PDF produced
        for node in document.traverse(condition=nodes.reference):
            refname = node.get('wiki_name')
            if refname is None:
                # Regular link: point back to the site
                refuri = node.get('refuri')
                if refuri is not None:
                    reference = get_reference(refuri.encode('utf_8'))
                    if isinstance(reference, Mailto):
                        # mailto:
                        node['refuri'] = str(reference)
                    else:
                        # Make canonical URI to the website for future
                        # download
                        node['refuri'] = str(context.uri.resolve(reference))
                continue
            # Now consider the link is valid
            title = node['name']
            document.nameids[title.lower()] = checkid(title)
            # The journey ends here for broken links
            if refname is False:
                continue
            # We extend the main page with the first level of subpages
            destination = resource.get_resource(refname)
            abspath = destination.get_abspath()
            if abspath not in pages:
                from page import WikiPage
                if not isinstance(destination, WikiPage):
                    # Link to a file, set the URI to download it
                    prefix = context.resource.get_pathto(destination)
                    node['refuri'] = str(context.uri.resolve(prefix))
                    continue
                # Adding the page to this one
                subdoc = destination.get_document()
                # We point the second level of links to the website
                for node in subdoc.traverse(condition=nodes.reference):
                    refname = node.get('wiki_name')
                    if refname is None:
                        # Regular link: point back to the site
                        refuri = node.get('refuri')
                        if refuri is not None:
                            reference = get_reference(refuri.encode('utf_8'))
                            if isinstance(reference, Mailto):
                                # mailto:
                                node['refuri'] = str(reference)
                            else:
                                refuri = context.uri.resolve(reference)
                                node['refuri'] = str(refuri)
                        continue
                    # Now consider the link is valid
                    title = node['name']
                    refid = title.lower()
                    if refid not in document.nameids:
                        document.nameids[refid] = checkid(title)
                    # The journey ends here for broken links
                    if refname is False:
                        continue
                    destination = resource.get_resource(refname)
                    prefix = context.resource.get_pathto(destination)
                    node['refuri'] = str(context.uri.resolve(prefix))
                # Now include the page
                # A page may begin with a section or a list of sections
                if len(subdoc) and isinstance(subdoc[0], nodes.section):
                    for node in subdoc.children:
                        if isinstance(node, nodes.section):
                            document.append(node)
                else:
                    subtitle = subdoc.get('title', u'')
                    section = nodes.section(*subdoc.children,
                                            **subdoc.attributes)
                    section.insert(0, nodes.title(text=subtitle))
                    document.append(section)
                pages.append(abspath)

        # Find the list of images to append
        for node in document.traverse(condition=nodes.image):
            uri = node['uri'].encode('utf_8')

            # Hack to handle local images
            reference = get_reference(uri)
            if not reference.scheme and uri.endswith('/;download'):
                reference = get_reference(uri[:-len('/;download')])

            if reference.scheme or reference.authority:
                # Fetch external image
                try:
                    image = get_handler(reference)
                    filename = reference.path.get_name()
                    name, ext, lang = FileName.decode(filename)
                    # At least images from ikaaro won't have an extension
                    if ext is None:
                        mimetype = vfs.get_mimetype(reference)
                        ext = guess_extension(mimetype)[1:]
                        filename = FileName.encode((name, ext, lang))
                except LookupError:
                    image = None
            else:
                try:
                    image = parent.get_resource(reference.path)
                    filename = image.get_property('filename')
                except LookupError:
                    image = None
            if image is None:
                # Missing image, prevent pdfLaTeX failure
                image = resource.get_resource('/ui/wiki/missing.png')
                filename = image.name
                # Remove all path so the image is found in tempdir
                node['uri'] = filename
                images.append((image, filename))
            else:
                # pdflatex does not support the ".jpeg" extension
                name, ext, lang = FileName.decode(filename)
                if ext == 'jpeg':
                    filename = FileName.encode((name, 'jpg', lang))
                # Remove all path so the image is found in tempdir
                node['uri'] = filename
                images.append((image, filename))

        # Make some modifications
        overrides = dict(resource.overrides)
        overrides['stylesheet'] = 'style.tex'
        output = publish_from_doctree(document, writer_name='latex',
                                      settings_overrides=overrides)
        output = figure_style_converter.sub(r'\\begin{figure}[H]', output)

        dirname = mkdtemp('wiki', 'itools')
        tempdir = vfs.open(dirname)

        # Save the document...
        file = tempdir.make_file(resource.name)
        try:
            file.write(output)
        finally:
            file.close()
        # The stylesheet...
        stylesheet = resource.get_resource('/ui/wiki/style.tex')
        file = tempdir.make_file('style.tex')
        try:
            stylesheet.save_state_to_file(file)
        finally:
            file.close()
        # The 'powered' image...
        image = resource.get_resource('/ui/aruni/images/ikaaro_powered.png')
        file = tempdir.make_file('ikaaro.png')
        try:
            image.save_state_to_file(file)
        finally:
            file.close()
        # And referenced images
        for image, filename in images:
            if tempdir.exists(filename):
                continue
            file = tempdir.make_file(filename)
            try:
                if isinstance(image, FileHandler):
                    try:
                        image.save_state_to_file(file)
                    except XMLError:
                        # XMLError is raised by unexpected HTTP responses
                        # from external images. See bug #249
                        pass
                else:
                    image.handler.save_state_to_file(file)
            finally:
                file.close()

        # From latex to PDF
        command = ['pdflatex', '-8bit', '-no-file-line-error',
                   '-interaction=batchmode', resource.name]
        try:
            call(command, cwd=dirname)
            # Twice for correct page numbering
            call(command, cwd=dirname)
        except OSError:
            msg = ERROR(u"PDF generation failed. Please install pdflatex.")
            return context.come_back(msg)

        pdfname = '%s.pdf' % resource.name
        if not tempdir.exists(pdfname):
            # TODO Print an error message somewhere with the 'dirname' for
            # inspection of the problem.
            return context.come_back(MSG(u"PDF generation failed."))

        # Read the file's data
        file = tempdir.open(pdfname)
        try:
            data = file.read()
        finally:
            file.close()

        # Clean the temporary folder
        vfs.remove(dirname)

        # Ok
        response = context.response
        response.set_header('Content-Type', 'application/pdf')
        response.set_header('Content-Disposition',
                            'attachment; filename=%s' % pdfname)
        return data



class WikiPage_Edit(STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Edit Page')
    template = '/ui/wiki/edit.xml'
    schema = {
        'title': Unicode,
        'data': String,
        'timestamp': DateTime,
    }


    def get_namespace(self, resource, context):
        context.styles.append(
                '/ui/tiny_mce/themes/advanced/skins/default/ui.css')
        context.styles.append('/ui/wiki/style.css')
        context.scripts.append('/ui/tiny_mce/tiny_mce_src.js')
        context.scripts.append('/ui/wiki/javascript.js')

        data = context.get_form_value('data') or resource.handler.to_str()
        return {
            'title': resource.get_title(),
            'data': data,
            'timestamp': DateTime.encode(datetime.now()),
        }


    def action(self, resource, context, form):
        if is_edit_conflict(resource, context, form['timestamp']):
            return

        title = form['title']
        language = resource.get_content_language(context)
        resource.set_property('title', title, language=language)

        # Data is assumed to be encoded in UTF-8
        data = form['data']
        # Save even if broken
        resource.handler.load_state_from_string(data)

        # Warn about syntax errors
        message = None
        try:
            html = resource.view.GET(resource, context)
            # Non-critical
            if 'class="system-message"' in html:
                message = ERROR(u"Syntax error, please check the view for "
                                u"details.")
        except SystemMessage, e:
            # Critical
            message = ERROR(u'Syntax error: $error', error=e.message)
        if message is None:
            accept = context.accept_language
            time = format_datetime(datetime.now(), accept=accept)
            message = messages.MSG_CHANGES_SAVED2(time=time)

        # Come back to the desired view
        if context.has_form_value('view'):
            goto = context.come_back(message)
            query = goto.query
            goto = goto.resolve(';view')
            goto.query = query
            return goto

        context.message = message



class WikiPage_Help(STLView):

    access = 'is_allowed_to_view'
    title = MSG(u"Help")
    template = '/ui/wiki/help.xml'


    def get_namespace(self, resource, context):
        context.styles.append('/ui/wiki/style.css')

        source = resource.get_resource('/ui/wiki/help.txt')
        source = source.to_str()

        overrides = dict(resource.overrides)
        overrides['stylesheet'] = ''
        html = publish_string(source, writer_name='html',
                               settings_overrides=overrides)
        document = XHTMLFile(string=html)
        events = document.get_body().get_content_elements()

        # Now remove some magic to make the help work like a wiki page
        source = source.split('.. XXX SPLIT HERE')[0]

        return {
            'help_source': source,
            'help_html': events,
        }



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Luis Arturo Belmar-Letelier <luis@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2008 Gautier Hayoun <gautier.hayoun@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import date, datetime, time
from re import compile
from textwrap import TextWrapper

# Import from itools
from itools.datatypes import Boolean, Date, Integer, String, Unicode
from itools.datatypes import XMLContent
from itools.gettext import MSG
from itools.ical import Time
from itools.html import xhtml_uri
from itools.i18n import format_datetime
from itools.web import STLForm, STLView
from itools.xml import XMLParser, START_ELEMENT, END_ELEMENT, TEXT

# Import from ikaaro
from ikaaro.messages import MSG_CHANGES_SAVED
from ikaaro.table_views import Table_View
from ikaaro.views import CompositeForm
from ikaaro.views import ContextMenu

# Local import
from datatypes import get_issue_fields, UsersList


###########################################################################
# Utilities
###########################################################################
url_expr = compile('([fh]t?tps?://[\w;/?:@&=+$,.#\-%]*)')
class OurWrapper(TextWrapper):

    def _split(self, text):
        # Override default's '_split' method to define URLs as unbreakable,
        # and reduce URLs if needed.
        # XXX This is fragile, since it uses TextWrapper private API.

        # Keep a mapping from reduced URL to full URL
        self.urls_map = {}

        # Get the chunks
        chunks = []
        for segment in url_expr.split(text):
            starts = segment.startswith
            if starts('http://') or starts('https://') or starts('ftp://'):
                if len(segment) > 70:
                    # Reduce URL
                    url = segment
                    segment = segment[:34] + '...' + segment[-33:]
                    self.urls_map[segment] = url
                else:
                    self.urls_map[segment] = segment
                chunks.append(segment)
            else:
                chunks.extend(TextWrapper._split(self, segment))
        return chunks



def indent(text):
    """Replace URLs by HTML links.  Wrap lines (with spaces) to 95 chars.
    """
    text = text.encode('utf-8')
    # Wrap
    buffer = []
    text_wrapper = OurWrapper(95)
    for line in text.splitlines():
        line = text_wrapper.fill(line) + '\n'
        for segment in url_expr.split(line):
            url = text_wrapper.urls_map.get(segment)
            if url is None:
                buffer.append(segment)
            else:
                if buffer:
                    yield TEXT, ''.join(buffer), 1
                    buffer = []
                # <a>...</a>
                attributes = {(None, 'href'): url}
                yield START_ELEMENT, (xhtml_uri, 'a', attributes), 1
                yield TEXT, segment, 1
                yield END_ELEMENT, (xhtml_uri, 'a'), 1
    if buffer:
        yield TEXT, ''.join(buffer), 1
        buffer = []



###########################################################################
# Menu
###########################################################################
class IssueTrackerMenu(ContextMenu):

    title = MSG(u'Tracker')

    def get_items(self, resource, context):
        items = [
            {'title': MSG(u'Search for issues'),
             'href': '%s/;search' % context.get_link(resource.parent)},
            {'title': MSG(u'Add a new issue'),
             'href': '%s/;add_issue' % context.get_link(resource.parent)}]
        return items



###########################################################################
# Views
###########################################################################
class Issue_Edit(STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Edit Issue')
    icon = 'edit.png'
    template = '/ui/tracker/edit_issue.xml'


    def get_schema(self, resource, context):
        return get_issue_fields(resource.parent)


    def get_value(self, resource, context, name, datatype):
        history = resource.get_history()
        record = history.get_record(-1)
        return  record.get_value(name)


    def get_namespace(self, resource, context):
        # Set Style & JS
        context.styles.append('/ui/tracker/tracker.css')
        context.scripts.append('/ui/tracker/tracker.js')

        # Local variables
        users = resource.get_resource('/users')
        history = resource.get_history()
        record = history.get_record(-1)

        # Build the namespace
        namespace = self.build_namespace(resource, context)

        # Comments
        comments = []
        i = 0
        for record in resource.get_history_records():
            comment = record.comment
            file = record.file
            if not comment and not file:
                continue
            rdatetime = record.datetime
            # solid in case the user has been removed
            username = record.username
            user_title = username
            if users.has_resource(username):
                user_title = users.get_resource(username).get_title()
            i += 1
            comments.append({
                'number': i,
                'user': user_title,
                'datetime': format_datetime(rdatetime),
                'comment': indent(comment),
                'file': file})
        comments.reverse()
        namespace['comments'] = comments

        # cc_list / cc_add / cc_remove
        cc_list = record.get_value('cc_list') or ()
        namespace['cc_list']= {'name': 'cc_list',
                              'value': [],
                              'class': None}
        namespace['cc_add']= {'name': 'cc_add',
                              'value': [],
                              'class': None}
        cc_value = namespace['cc_list']['value']
        add_value = namespace['cc_add']['value']
        for user in UsersList(tracker=resource.parent).get_options():
            user['selected'] = False
            if user['name'] in cc_list:
                cc_value.append(user)
            else:
                add_value.append(user)
        namespace['cc_remove'] = None

        # Reported by
        reported_by = resource.get_reported_by()
        namespace['reported_by'] = users.get_resource(reported_by).get_title()

        # list_products
        tracker = resource.parent
        namespace['list_products'] = tracker.get_list_products_namespace()

        return namespace


    def action(self, resource, context, form):
        # Edit
        resource._add_record(context, form)
        # Change
        context.server.change_resource(resource)
        context.message = MSG_CHANGES_SAVED



class Issue_History(STLView):

    access = 'is_allowed_to_view'
    title = MSG(u'History')
    icon = 'history.png'
    template = '/ui/tracker/issue_history.xml'


    def get_namespace(self, resource, context):
        # Set Style
        context.styles.append('/ui/tracker/tracker.css')

        # Local variables
        users = resource.get_resource('/users')
        tracker = resource.parent
        versions = tracker.get_resource('version').handler
        types = tracker.get_resource('type').handler
        states = tracker.get_resource('state').handler
        modules = tracker.get_resource('module').handler
        priorities = tracker.get_resource('priority').handler
        # Initial values
        previous_title = None
        previous_version = None
        previous_type = None
        previous_state = None
        previous_module = None
        previous_priority = None
        previous_assigned_to = None
        previous_cc_list = None

        # Build the namespace
        rows = []
        i = 0
        for record in resource.get_history_records():
            rdatetime = record.get_value('datetime')
            username = record.get_value('username')
            title = record.get_value('title')
            module = record.get_value('module')
            version = record.get_value('version')
            type = record.get_value('type')
            priority = record.get_value('priority')
            assigned_to = record.get_value('assigned_to')
            state = record.get_value('state')
            comment = record.get_value('comment')
            cc_list = record.get_value('cc_list') or ()
            file = record.get_value('file')
            # Solid in case the user has been removed
            user_exist = users.has_resource(username)
            usertitle = (user_exist and
                         users.get_resource(username).get_title() or username)
            comment = XMLContent.encode(Unicode.encode(comment))
            comment = XMLParser(comment.replace('\n', '<br />'))
            i += 1
            row_ns = {'number': i,
                      'user': usertitle,
                      'datetime': format_datetime(rdatetime),
                      'title': None,
                      'version': None,
                      'type': None,
                      'state': None,
                      'module': None,
                      'priority': None,
                      'assigned_to': None,
                      'comment': comment,
                      'cc_list': None,
                      'file': file}

            if title != previous_title:
                previous_title = title
                row_ns['title'] = title
            if version != previous_version:
                previous_version = version
                row_ns['version'] = ' '
                if module is not None:
                    version = versions.get_record(int(version))
                    if version:
                        value = versions.get_record_value(version, 'title')
                        row_ns['version'] = value
            if type != previous_type:
                previous_type = type
                row_ns['type'] = ' '
                if type is not None:
                    type = types.get_record(int(type))
                    if type is not None:
                        value = types.get_record_value(type, 'title')
                        row_ns['type'] = value
            if state != previous_state:
                previous_state = state
                row_ns['state'] = ' '
                if state is not None:
                    state = states.get_record(int(state))
                    if state is not None:
                        value = states.get_record_value(state, 'title')
                        row_ns['state'] = value
            if module != previous_module:
                previous_module = module
                row_ns['module'] = ' '
                if module is not None:
                    module = modules.get_record(int(module))
                    if module is not None:
                        value = modules.get_record_value(module, 'title')
                        row_ns['module'] = value
            if priority != previous_priority:
                previous_priority = priority
                row_ns['priority'] = ' '
                if priority is not None:
                    priority = priorities.get_record(int(priority))
                    if priority is not None:
                        value = priorities.get_record_value(priority, 'title')
                        row_ns['priority'] = value
            if assigned_to != previous_assigned_to:
                previous_assigned_to = assigned_to
                if assigned_to and users.has_resource(assigned_to):
                    assigned_to_user = users.get_resource(assigned_to)
                    row_ns['assigned_to'] = assigned_to_user.get_title()
                else:
                    row_ns['assigned_to'] = ' '
            if cc_list != previous_cc_list:
                root = context.root
                previous_cc_list = cc_list
                new_values = []
                for cc in cc_list:
                    user = root.get_user(cc)
                    if user:
                        new_values.append(user.get_property('email'))
                if new_values:
                    row_ns['cc_list'] = u', '.join(new_values)
                else:
                    row_ns['cc_list'] = ' '

            rows.append(row_ns)

        rows.reverse()

        # Ok
        return {'number': resource.name, 'rows': rows}



class Issue_ViewResources(Table_View):

    search_template = None

    batch_msg1 = MSG(u"There is 1 assignment.")
    batch_msg2 = MSG(u"There are ${n} assignments.")


    def get_items(self, resource, context):
        issue = context.resource
        return resource.handler.search(issue=issue.name)


    def get_item_value(self, resource, context, item, column):
        if column == 'id':
            id = item.id
            return id, ';edit_resources?id=%s' % id
        return Table_View.get_item_value(self, resource, context, item,
                                         column)


    def action_remove(self, resource, context, form):
        calendar = resource.get_calendar()
        Table_View.action_remove(self, calendar, context, form)



class Issue_AddEditResource(STLForm):

    access = 'is_allowed_to_edit'
    template = '/ui/tracker/edit_resource.xml'

    query_schema = {
        'id': Integer}

    schema = {
        'resource': String,
        'dtstart': Date(mandatory=True),
        'dtend': Date(mandatory=True),
        'tstart': Time(default=time(0, 0)),
        'tend': Time(default=time(0, 0)),
        'comment': Unicode}


    def get_namespace(self, resource, context):
        calendar = resource.get_calendar()

        # Add or Edit
        id = context.query['id']
        if id is None:
            # Add a new resource-issue
            action = ';edit_resources'
            user = None
            d_start = d_end = date.today()
            t_start = t_end = ''
            comment = u''
        else:
            # Edit a resource-issue
            action = ';edit_resources?id=%s' % id
            record = calendar.handler.get_record(id)
            get_value = calendar.handler.get_record_value
            user = get_value(record, 'resource')
            dtstart = get_value(record, 'dtstart')
            dtend = get_value(record, 'dtend')
            comment = get_value(record, 'comment')
            d_start, t_start = dtstart.date(), dtstart.time()
            d_end, t_end = dtend.date(), dtend.time()
            t_start = t_start.strftime('%H:%M')
            t_end = t_end.strftime('%H:%M')
            id = str(id)

        # Time select
        timetables = calendar.get_timetables()
        time_select = [
            {'name': index,
             'start': start.strftime('%H:%M'),
             'end': end.strftime('%H:%M')}
            for index, (start, end) in enumerate(timetables) ]

        # Ok
        return {
            'action': action,
            'id': id,
            'users': resource.parent.get_members_namespace(user),
            'd_start': d_start.strftime('%Y-%m-%d'),
            't_start': t_start,
            'd_end': d_end.strftime('%Y-%m-%d'),
            't_end': t_end,
            'time_select': time_select,
            'comment': comment}


    def action_add(self, resource, context, form):
        dtstart = datetime.combine(form['dtstart'], form['tstart'])
        dtend = datetime.combine(form['dtend'], form['tend'])
        record = {
            'issue': resource.name,
            'resource': form['resource'],
            'dtstart': dtstart,
            'dtend': dtend,
            'comment': form['comment']}

        # Change
        calendar = resource.get_calendar()
        calendar.handler.add_record(record)
        # Ok
        context.message = MSG_CHANGES_SAVED


    def action_edit(self, resource, context, form):
        id = context.query['id']

        # New record
        dtstart = datetime.combine(form['dtstart'], form['tstart'])
        dtend = datetime.combine(form['dtend'], form['tend'])
        record = {
            'issue': resource.name,
            'resource': form['resource'],
            'dtstart': dtstart,
            'dtend': dtend,
            'comment': form['comment']}

        # Change
        calendar = resource.get_calendar()
        calendar.handler.update_record(id, **record)
        # Ok
        context.message = MSG_CHANGES_SAVED



class Issue_EditResources(CompositeForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Edit resources')
    icon = 'edit.png'

    subviews = [
        Issue_AddEditResource(),
        Issue_ViewResources(),
    ]

    def get_namespace(self, resource, context):
        # Override so we can pass a different resource to Issue_ViewResources
        calendar = resource.get_calendar()
        views = [
            self.subviews[0].GET(resource, context),
            self.subviews[1].GET(calendar, context) ]

        return {'views': views}



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from optparse import OptionParser
from os import mkdir
from string import Template
import sys

# Import from itools
import itools
from itools.handlers import Database
from itools.uri import get_absolute_reference
from itools.xapian import make_catalog, CatalogAware

# Import from ikaaro
from ikaaro.root import Root
from ikaaro.utils import generate_password


template = Template(
"""# The "modules" variable lists the Python modules or packages that will be
# loaded when the applications starts.
#
modules = ${modules}

# The "listen-address" and "listen-port" variables define, respectively, the
# internet address and the port number the web server listens to for HTTP
# connections.
#
# By default connections are accepted from any internet address.  And the
# server listens the 8080 port number.
#
listen-address = ${listen_address}
listen-port = ${listen_port}

# The "smtp-host" variable defines the name or IP address of the SMTP relay.
# The "smtp-from" variable is the email address used in the From field when
# sending anonymous emails.  (These options are required for the application
# to send emails).
#
# The "smtp-login" and "smtp-password" variables define the credentials
# required to access a secured SMTP server.
#
smtp-host = ${smtp_host}
smtp-from = ${smtp_from}
smtp-login =
smtp-password =

# The "debug" variable defines whether the web server will be run in debug
# mode or not.  When run in debug mode (debug = 1), debugging information
# will be written to the "log/events" file.  By default debug mode is not
# active (debug = 0).
#
debug = 0

# The "index-text" vairable defines whether the catalog must process full-text
# indexing. It requires (much) more time and third-party applications.
# To speed up catalog updates, set this option to 0 (default is 1).
index-text = 1
""")



def init(parser, options, target):
    try:
        mkdir(target)
    except OSError:
        parser.error('can not create the instance (check permissions)')

    # Get the email address for the init user
    if options.email is None:
        sys.stdout.write("Type your email address: ")
        email = sys.stdin.readline().strip()
    else:
        email = options.email

    # Get the password
    if options.password is None:
        password = generate_password()
    else:
        password = options.password

    # Load the root class
    if options.root is None:
        root_class = Root
        modules = ''
    else:
        modules = options.root
        exec('import %s' % modules)
        exec('root_class = %s.Root' % modules)

    # The configuration file
    namespace = {
        'modules': modules,
        'listen_address': getattr(options, 'address') or '',
        'listen_port': getattr(options, 'port') or '8080',
        'smtp_host': getattr(options, 'smtp_host') or 'localhost',
        'smtp_from': email,
    }
    config = template.substitute(**namespace)
    open('%s/config.conf' % target, 'w').write(config)

    # Create the folder structure
    mkdir('%s/database' % target)
    mkdir('%s/log' % target)
    mkdir('%s/spool' % target)
    # Make the root
    database = Database()
    base = get_absolute_reference(target).resolve2('database')
    folder = database.get_handler(base)
    root = root_class._make_resource(root_class, folder, email, password)
    database.save_changes()
    # Index everything
    catalog = make_catalog('%s/catalog' % target)
    for handler in root.traverse_resources():
        if isinstance(handler, CatalogAware):
            catalog.index_document(handler)
    catalog.save_changes()

    # Bravo!
    print '*'
    print '* Welcome to ikaaro'
    print '* A user with administration rights has been created for you:'
    print '*   username: %s' % email
    print '*   password: %s' % password
    print '*'
    print '* To start the new instance type:'
    print '*   icms-start.py %s' % target
    print '*'



if __name__ == '__main__':
    # The command line parser
    usage = '%prog [OPTIONS] TARGET'
    version = 'itools %s' % itools.__version__
    description = 'Creates a new instance of ikaaro with the name TARGET.'
    parser = OptionParser(usage, version=version, description=description)
    parser.add_option('-a', '--address', help='listen to IP ADDRESS')
    parser.add_option('-e', '--email', help='e-mail address of the admin user')
    parser.add_option('-p', '--port', type='int', help='listen to PORT number')
    parser.add_option('-r', '--root',
        help='create an instance of the ROOT application')
    parser.add_option('-s', '--smtp-host',
        help='use the given SMTP_HOST to send emails')
    parser.add_option('-w', '--password',
        help='use the given PASSWORD for the admin user')

    options, args = parser.parse_args()
    if len(args) != 1:
        parser.error('incorrect number of arguments')

    target = args[0]

    # Action!
    init(parser, options, target)



# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.datatypes import Boolean, String
from itools.gettext import MSG
from itools.i18n import get_language_name, get_languages
from itools.uri import Path
from itools.web import STLView, STLForm, INFO, ERROR
from itools.xapian import PhraseQuery, AndQuery

# Import from ikaaro
from access import RoleAware_BrowseUsers, RoleAware_AddUser
from access import RoleAware_EditMembership
from folder_views import Folder_Orphans
import messages
from utils import get_base_path_query
from views import IconsView, ContextMenu



###########################################################################
# The menu
###########################################################################
class ControlPanelMenu(ContextMenu):

    title = MSG(u'Control Panel')

    def get_items(self, resource, context):
        items = []
        for name in resource.class_control_panel:
            view = resource.get_view(name)
            if view is None:
                continue
            if not resource.is_access_allowed(context.user, resource, view):
                continue
            items.append({
                'title': view.title,
                'src': resource.get_method_icon(view, size='16x16'),
                'href': ';%s' % name})

        return items



class CPBaseView(object):
    context_menus = [ControlPanelMenu()]



###########################################################################
# Views
###########################################################################
class ControlPanel(CPBaseView, IconsView):

    access = 'is_allowed_to_edit'
    title = MSG(u'Control Panel')
    icon = 'settings.png'


    def get_namespace(self, resource, context):
        items = []
        for name in resource.class_control_panel:
            view = resource.get_view(name)
            if view is None:
                continue
            if not resource.is_access_allowed(context.user, resource, view):
                continue
            items.append({
                'icon': resource.get_method_icon(view, size='48x48'),
                'title': view.title,
                'description': view.description,
                'url': ';%s' % name})

        return {
            'title': MSG(u'Control Panel'),
            'batch': None,
            'items': items}



class CPEditVirtualHosts(CPBaseView, STLForm):

    access = 'is_admin'
    title = MSG(u'Virtual Hosts')
    icon = 'website.png'
    description = MSG(u'Define the domain names for this Web Site.')
    template = '/ui/website/virtual_hosts.xml'
    schema = {
        'vhosts': String}


    def get_namespace(self, resource, context):
        vhosts = resource.get_property('vhosts')
        return {
            'vhosts': '\n'.join(vhosts)}


    def action(self, resource, context, form):
        vhosts = form['vhosts']
        vhosts = [ x.strip() for x in vhosts.splitlines() ]
        vhosts = [ x for x in vhosts if x ]
        vhosts = tuple(vhosts)
        resource.set_property('vhosts', vhosts)
        # Ok
        context.message = messages.MSG_CHANGES_SAVED



class CPEditSecurityPolicy(CPBaseView, STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Security Policy')
    icon = 'lock.png'
    description = MSG(u'Choose the security policy.')
    template = '/ui/website/security_policy.xml'
    schema = {
        'website_is_open': Boolean(default=False)}


    def get_namespace(self, resource, context):
        is_open = resource.get_property('website_is_open')
        return {
            'is_open': is_open,
            'is_closed': not is_open}


    def action(self, resource, context, form):
        value = form['website_is_open']
        resource.set_property('website_is_open', value)
        # Ok
        context.message = messages.MSG_CHANGES_SAVED



class CPEditContactOptions(CPBaseView, STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Contact Options')
    icon = 'mail.png'
    description = MSG(u'Configure the Contact form.')
    template = '/ui/website/contact_options.xml'
    schema = {
        'contacts': String(multiple=True)}


    def get_namespace(self, resource, context):
        # Find out the contacts
        contacts = resource.get_property('contacts')

        # Build the namespace
        users = resource.get_resource('/users')
        # Only members of the website are showed
        namespace = {}
        namespace['contacts'] = []
        for username in resource.get_members():
            user = users.get_resource(username)
            email = user.get_property('email')
            if not email:
                continue
            namespace['contacts'].append(
                {'name': username,
                 'email': email,
                 'title': user.get_title(),
                 'is_selected': username in contacts})

        # Sort
        namespace['contacts'].sort(key=lambda x: x['email'])

        return namespace


    def action(self, resource, context, form):
        contacts = form['contacts']
        contacts = tuple(contacts)
        resource.set_property('contacts', contacts)
        # Ok
        context.message = messages.MSG_CHANGES_SAVED



class CPBrokenLinks(CPBaseView, STLView):

    access = 'is_admin'
    title = MSG(u'Broken Links')
    icon = 'clear.png'
    description = MSG(u'Check the referential integrity.')
    template = '/ui/website/broken_links.xml'


    def get_namespace(self, resource, context):
        # Find out broken links
        catalog = context.server.catalog
        base = resource.get_abspath()

        # Search only within the given resource
        base_str = str(base)
        query = get_base_path_query(base_str, include_container=True)
        results = catalog.search(query)

        # Find out the broken links
        root = context.root
        broken = {}
        for link in catalog.get_unique_values('links'):
            if root.has_resource(link):
                continue
            sub_results = results.search(PhraseQuery('links', link))
            link = str(base.get_pathto(Path(link)))
            for brain in sub_results.get_documents():
                broken.setdefault(brain.abspath, []).append(link)

        # Build the namespace
        items = []
        total = 0
        keys = broken.keys()
        keys.sort()
        for path in keys:
            links = broken[path]
            path = str(base.get_pathto(Path(path)))
            n = len(links)
            items.append({'path': path, 'links': links, 'n': n})
            total += n

        return {
            'items': items,
            'total': total}



class CPEditLanguages(CPBaseView, STLForm):

    access = 'is_admin'
    title = MSG(u'Languages')
    description = MSG(u'Define the Web Site languages.')
    icon = 'languages.png'
    template = '/ui/website/edit_languages.xml'
    schema = {
        'codes': String(multiple=True, mandatory=True)}


    def get_namespace(self, resource, context):
        ws_languages = resource.get_property('website_languages')

        # Active languages
        default = ws_languages[0]
        active = []
        for code in ws_languages:
            language_name = get_language_name(code)
            active.append({
                'code': code,
                'name': language_name,
                'isdefault': code == default})

        # Not active languages
        not_active = [
            x for x in get_languages() if x['code'] not in ws_languages ]
        not_active.sort(lambda x, y: cmp(x['name'], y['name']))

        # Ok
        return {
            'active_languages': active,
            'not_active_languages': not_active}


    #######################################################################
    # Actions / Edit
    def action_change_default_language(self, resource, context, form):
        codes = form['codes']

        # This action requires only one language to be selected
        if len(codes) != 1:
            message = ERROR(u'You must select one and only one language.')
            context.message = message
            return
        default = codes[0]

        # Change the default language
        languages = resource.get_property('website_languages')
        languages = [ x for x in languages if x != default ]
        languages.insert(0, default)
        resource.set_property('website_languages', tuple(languages))
        # Ok
        context.message = INFO(u'The default language has been changed.')


    def action_remove_languages(self, resource, context, form):
        codes = form['codes']

        # Check the default language is not to be removed
        languages = resource.get_property('website_languages')
        default = languages[0]
        if default in codes:
            message = ERROR(u'You can not remove the default language.')
            context.message = message
            return

        # Remove the languages
        languages = [ x for x in languages if x not in codes ]
        resource.set_property('website_languages', tuple(languages))
        # Ok
        context.message = INFO(u'Languages removed.')


    #######################################################################
    # Actions / Add
    action_add_language_schema = {
        'code': String(mandatory=True)}

    def action_add_language(self, resource, context, form):
        code = form['code']

        ws_languages = resource.get_property('website_languages')
        resource.set_property('website_languages', ws_languages + (code,))
        # Ok
        context.message = INFO(u'Language added.')




###########################################################################
# Add the control panel menu to views defined somewhere else
###########################################################################
class CPBrowseUsers(CPBaseView, RoleAware_BrowseUsers):
    pass


class CPAddUser(CPBaseView, RoleAware_AddUser):
    pass


class CPEditMembership(CPBaseView, RoleAware_EditMembership):
    pass


class CPOrphans(CPBaseView, Folder_Orphans):
    pass




# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.datatypes import DataType, String, Unicode, XMLContent
from itools.handlers import File, register_handler_class
from itools.utils import freeze, add_type
from itools.web import get_context
from itools.xml import xml_uri, XMLParser, START_ELEMENT, END_ELEMENT, TEXT

# Import from ikaaro
from exceptions import ParserError
from registry import get_resource_class



class Record(DataType):

    # Set your own default list to avoid sharing this instance
    default = freeze([])
    schema = {}

    multiple = True



def get_datatype(format, name):
    if format is None:
        return String

    cls = get_resource_class(format)
    if cls is None:
        return String

    schema = cls.get_metadata_schema()
    return schema.get(name, String)



class Metadata(File):

    class_mimetypes = ['text/x-metadata']
    class_extension = 'metadata'


    def new(self, handler_class=None, format=None, **kw):
        # Add format and version
        self.format = format or handler_class.class_id
        self.version = handler_class.class_version

        # Initialize
        properties = {}
        for name in kw:
            value = kw[name]
            if value is not None:
                properties[name] = value

        # Set state
        self.properties = properties


    def _load_state_from_file(self, file):
        # Constants
        xml_lang = (xml_uri, 'lang')
        error1 = 'unexpected start tag "%s" at line "%s"'
        # Local variables
        language = None
        stack = []

        # Parse
        for type, value, line in XMLParser(file.read()):
            if type == START_ELEMENT:
                ns_uri, name, attributes = value
                # First tag: <metadata>
                n = len(stack)
                if n == 0:
                    if name != 'metadata':
                        raise ParserError, error1 % (name, line)
                    self.format = attributes.get((None, 'format'))
                    self.version = attributes.get((None, 'version'))
                    if self.format is None:
                        schema = {}
                    else:
                        cls = get_resource_class(self.format)
                        schema = cls.get_metadata_schema()
                    stack.append((name, None, {}))
                    continue

                # Find out datatype
                if n == 1:
                    datatype = schema.get(name, String)
                else:
                    datatype = stack[-1][1]
                    if issubclass(datatype, Record):
                        datatype = datatype.schema.get(name, String)
                    else:
                        raise ParserError, error1 % (name, line)

                if issubclass(datatype, Record):
                    stack.append((name, datatype, {}))
                else:
                    stack.append((name, datatype, ''))
                    language = attributes.get(xml_lang)
            elif type == END_ELEMENT:
                name, datatype, value = stack.pop()

                # Last tag: </metadata>
                n = len(stack)
                if n == 0:
                    self.properties = value
                    break

                # Decode value
                if issubclass(datatype, Record):
                    pass
                elif issubclass(datatype, Unicode):
                    value = datatype.decode(value, 'utf-8')
                else:
                    value = datatype.decode(value)

                # Set property
                is_multiple = getattr(datatype, 'multiple', False)
                if is_multiple:
                    stack[-1][2].setdefault(name, []).append(value)
                elif language is None:
                    stack[-1][2][name] = value
                else:
                    stack[-1][2].setdefault(name, {})
                    stack[-1][2][name][language] = value
                # Reset variables
                language = None
            elif type == TEXT:
                n = len(stack)
                if n == 0:
                    continue
                if isinstance(stack[-1][2], dict):
                    continue

                name, datatype, last_value = stack.pop()
                value = last_value + value
                stack.append((name, datatype, value))


    def to_str(self):
        # format, version, schema
        format = self.format
        version = self.version
        cls = get_resource_class(format)
        if cls is None:
            schema = {}
        else:
            schema = cls.get_metadata_schema()

        # Opening
        lines = ['<?xml version="1.0" encoding="UTF-8"?>\n',
                 '<metadata format="%s" version="%s">\n' % (format, version)]

        # Properties
        for name in self.properties:
            value = self.properties[name]
            datatype = schema.get(name, String)
            is_multiple = getattr(datatype, 'multiple', False)

            # Multilingual properties
            if isinstance(value, dict):
                template = '  <%s xml:lang="%s">%s</%s>\n'
                for language, value in value.items():
                    value = datatype.encode(value)
                    value = XMLContent.encode(value)
                    lines.append(template % (name, language, value, name))
            # Multiple values
            elif is_multiple:
                if not isinstance(value, list):
                    raise TypeError, 'multiple values must be lists'
                # Record
                if issubclass(datatype, Record):
                    aux = datatype.schema
                    for value in value:
                        lines.append('  <%s>\n' % name)
                        for key, value in value.items():
                            value = aux.get(key, String).encode(value)
                            value = XMLContent.encode(value)
                            lines.append('    <%s>%s</%s>\n'
                                         % (key, value, key))
                        lines.append('  </%s>\n' % name)
                    continue
                # Regular field
                for value in value:
                    value = datatype.encode(value)
                    value = XMLContent.encode(value)
                    lines.append('  <%s>%s</%s>\n' % (name, value, name))
                continue
            # Simple properties
            else:
                value = datatype.encode(value)
                value = XMLContent.encode(value)
                lines.append('  <%s>%s</%s>\n' % (name, value, name))

        lines.append('</metadata>\n')
        return ''.join(lines)


    ########################################################################
    # API
    ########################################################################
    def get_property_and_language(self, name, language=None):
        """Return the value for the given property and the language of that
        value.

        For monolingual properties, the language always will be None.
        """
        # Check the property exists
        datatype = get_datatype(self.format, name)
        if name not in self.properties:
            default = datatype.get_default()
            return default, None
        # Get the value
        value = self.properties[name]

        # Monolingual property
        if not isinstance(value, dict):
            return value, None

        # Language negotiation
        if language is None:
            context = get_context()
            if context is None:
                language = None
            else:
                languages = [
                    k for k, v in value.items() if not datatype.is_empty(v) ]
                accept = context.accept_language
                language = accept.select_language(languages)
            # Default (FIXME pick one at random)
            if language is None:
                language = value.keys()[0]
            return value[language], language

        if language in value:
            return value[language], language
        return datatype.get_default(), None


    def get_property(self, name, language=None):
        return self.get_property_and_language(name, language=language)[0]


    def has_property(self, name, language=None):
        if name not in self.properties:
            return False

        if language is not None:
            return language in self.properties[name]

        return True


    def set_property(self, name, value, language=None):
        self.set_changed()

        # Set the value
        if language is None:
            datatype = get_datatype(self.format, name)
            is_multiple = getattr(datatype, 'multiple', False)

            default = datatype.get_default()
            if is_multiple:
                if isinstance(value, list):
                    self.properties[name] = value
                else:
                    values = self.properties.setdefault(name, [])
                    values.append(value)
            else:
                self.properties[name] = value
        else:
            values = self.properties.setdefault(name, {})
            values[language] = value


    def del_property(self, name, language=None):
        if name in self.properties:
            if language is None:
                self.set_changed()
                del self.properties[name]
            else:
                values = self.properties[name]
                if language in values:
                    self.set_changed()
                    del values[language]



###########################################################################
# Register
###########################################################################
register_handler_class(Metadata)
for mimetype in Metadata.class_mimetypes:
    add_type(mimetype, '.%s' % Metadata.class_extension)




# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Luis Arturo Belmar-Letelier <luis@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime

# Import from itools
from itools.csv import CSVFile, Property
from itools.datatypes import Boolean, Integer, String, Unicode, Enumerate
from itools.gettext import MSG
from itools.i18n import format_datetime
from itools.handlers import merge_dicts
from itools.uri import encode_query, Reference
from itools.web import BaseView, BaseForm, STLForm
from itools.web import INFO, ERROR
from itools.web.views import process_form

# Import from ikaaro
from ikaaro.buttons import Button
from ikaaro.forms import HiddenWidget, TextWidget
from ikaaro import messages
from ikaaro.resource_views import DBResource_NewInstance
from ikaaro.views import BrowseForm, SearchForm as BaseSearchForm, ContextMenu

# Import from ikaaro.tracker
from issue import Issue
from datatypes import get_issue_fields, TrackerList, ProductInfoList
from datatypes import UsersList
from stored import StoredSearchFile, StoredSearch


columns = [
    ('id', MSG(u'Id')),
    ('title', MSG(u'Title')),
    ('product', MSG(u'Product')),
    ('module', MSG(u'Module')),
    ('version', MSG(u'Version')),
    ('type', MSG(u'Type')),
    ('state', MSG(u'State')),
    ('priority', MSG(u'Priority')),
    ('assigned_to', MSG(u'Assigned To')),
    ('mtime', MSG(u'Modified'))]



###########################################################################
# Menus
###########################################################################
class GoToIssueMenu(ContextMenu):

    title = MSG(u'Go To Issue')
    template = '/ui/tracker/menu_goto.xml'

    def get_namespace(self, resource, context):
        path_to_tracker = '..' if isinstance(resource, Issue) else '.'
        return {
            'path_to_tracker': path_to_tracker,
            'title': self.title}



class StoreSearchMenu(ContextMenu):
    """Form to store a search.
    """

    title = MSG(u'Remember this search')
    template = '/ui/tracker/menu_remember.xml'
    query_schema = merge_dicts(StoredSearchFile.schema,
                               search_name=String,
                               search_title=Unicode)

    def get_namespace(self, resource, context):
        # This search exists ?
        search_name = context.get_query_value('search_name')
        search = None
        if search_name:
            try:
                search = resource.get_resource(search_name)
            except LookupError:
                pass

        # Set the get function and the title
        if search:
            get = search.get_values
            search_title = search.get_title()
        else:
            # Warning, a menu is not the default view!
            query = process_form(context.get_query_value, self.query_schema)
            get = query.get
            search_title = None

        # Fill the fields
        fields = []
        for name, type in StoredSearchFile.schema.iteritems():
            value = get(name)
            if isinstance(value, list):
                for x in value:
                    fields.append({'name': name, 'value': type.encode(x)})
            elif value is not None:
                fields.append({'name': name, 'value': type.encode(value)})

        # Ok
        return {'title': self.title,
                'search_title': search_title,
                'search_name': search_name,
                'search_fields': fields }



class StoredSearchesMenu(ContextMenu):
    """Provides links to every stored search.
    """

    title = MSG(u'Stored Searches')

    def get_items(self, resource, context):
        # If called from a child
        if isinstance(resource, Issue):
            resource = resource.parent

        # Namespace
        search_name = context.get_query_value('search_name')
        base = '%s/;view' % context.get_link(resource)
        items = resource.search_resources(cls=StoredSearch)
        items = [
            {'title': x.get_property('title'),
             'href': '%s?search_name=%s' % (base, x.name),
             'class': 'nav_active' if (x.name == search_name) else None}
            for x in items ]
        items.sort(lambda x, y: cmp(x['title'], y['title']))

        return items



class TrackerViewMenu(ContextMenu):

    title = MSG(u'Advanced')

    def get_items(self, resource, context):
        # Keep the query parameters
        schema = context.view.get_query_schema()
        params = encode_query(context.query, schema)
        items = [
            {'title': MSG(u'Edit this search'),
             'href': ';search?%s' % params},
            {'title': MSG(u'Change Several Issues'),
             'href': ';change_several_bugs?%s' % params},
            {'title': MSG(u'Export to Text'),
             'href': ';export_to_text?%s' % params},
            {'title': MSG(u'Export to CSV'),
             'href': ';export_to_csv_form?%s' % params},
            {'title': MSG(u'Resources'),
             'href': 'calendar/;monthly_view?%s' % params}]
        return items



###########################################################################
# Views
###########################################################################
class Tracker_NewInstance(DBResource_NewInstance):

    schema = merge_dicts(
        DBResource_NewInstance.schema,
        product=Unicode(mandatory=True))

    widgets = DBResource_NewInstance.widgets + \
        [TextWidget('product', title=MSG(u'Give the title of one Product'))]


    def action(self, resource, context, form):
        ok = DBResource_NewInstance.action(self, resource, context, form)
        if ok is None:
            return

        # Add the initial product
        name = form['name']
        product = form['product']
        table = resource.get_resource('%s/product' % name).get_handler()
        product = Property(product, language='en')
        table.add_record({'title': product})

        # Ok
        return ok



class Tracker_AddIssue(STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Add')
    icon = 'new.png'
    template = '/ui/tracker/add_issue.xml'


    def get_schema(self, resource, context):
        return get_issue_fields(resource)


    def get_namespace(self, resource, context):
        context.styles.append('/ui/tracker/tracker.css')
        context.scripts.append('/ui/tracker/tracker.js')

        namespace =  self.build_namespace(resource, context)
        namespace['list_products'] = resource.get_list_products_namespace()

        return namespace


    def action(self, resource, context, form):
        # Add
        id = resource.get_new_id()
        issue = Issue.make_resource(Issue, resource, id)
        issue._add_record(context, form)

        # Ok
        message = INFO(u'New issue added.')
        goto = './%s/' % id
        return context.come_back(message, goto=goto)



class Tracker_View(BrowseForm):

    access = 'is_allowed_to_view'
    title = MSG(u'View')
    icon = 'view.png'

    schema = {
        'ids': String(multiple=True, mandatory=True)}

    tracker_schema = {
        # Do not batch
        'batch_size': Integer(default=0),
        # search_fields
        'search_name': Unicode(),
        'mtime': Integer(default=0),
        'product': Integer(multiple=True),
        'module': Integer(multiple=True),
        'version': Integer(multiple=True),
        'type': Integer(multiple=True),
        'state': Integer(multiple=True),
        'priority': Integer(multiple=True),
        'assigned_to': String(multiple=True),
        # Specific fields
        'search_field': String,
        'search_term': Unicode,
        'search_subfolders': Boolean(default=False),
        # BrowseForm fields
        'sort_by': String,
        'reverse': Boolean(default=None),
    }

    context_menus = [StoreSearchMenu(), TrackerViewMenu()]


    def get_query_schema(self):
        return merge_dicts(BrowseForm.get_query_schema(self),
                           self.tracker_schema)


    def GET(self, resource, context):
        # Check stored search
        search_name = context.query['search_name']
        if search_name:
            try:
                search = resource.get_resource(search_name)
            except LookupError:
                msg = MSG(u'Unknown stored search "${sname}".')
                goto = ';search'
                return context.come_back(msg, goto=goto, sname=search_name)
        # Ok
        return BrowseForm.GET(self, resource, context)


    def get_namespace(self, resource, context):
        # Set Style
        context.styles.append('/ui/tracker/tracker.css')

        # Default table namespace
        namespace = BrowseForm.get_namespace(self, resource, context)

        # Number of results
        query = context.query
        search_name = query['search_name']
        if search_name:
            search = resource.get_resource(search_name)

        # Keep the search_parameters, clean different actions
        schema = self.get_query_schema()
        namespace['search_parameters'] = encode_query(query, schema)

        return namespace


    def get_items(self, resource, context):
        return resource.get_search_results(context)


    def sort_and_batch(self, resource, context, results):
        query = context.query
        # Stored search, or default
        search_name = query['search_name']
        if search_name:
            search = resource.get_resource(search_name).handler
            sort_by = search.get_value('sort_by')
            reverse = search.get_value('reverse')
        else:
            sort_by = 'title'
            reverse = False
        # Query takes precedence
        if query['sort_by'] is not None:
            sort_by = query['sort_by']
        if query['reverse'] is not None:
            reverse = query['reverse']
        return results.get_documents(sort_by=sort_by, reverse=reverse)


    def get_item_value(self, resource, context, item, column):
        if column == 'checkbox':
            selected_issues = context.get_form_values('ids') or []
            return item.name, item.name in selected_issues
        if column == 'id':
            id = item.name
            return id, '%s/;edit' % id

        value = getattr(item, column)
        if value is None:
            return None
        if column == 'title':
            return value, '%s/;edit' % item.name
        # Assigned to
        if column == 'assigned_to':
            users = resource.get_resource('/users')
            try:
                user = users.get_resource(value)
            except LookupError:
                return None
            return user.get_title()
        # Mtime
        if column == 'mtime':
            mtime = datetime.strptime(value, '%Y%m%d%H%M%S')
            return format_datetime(mtime)

        # Tables
        table = resource.get_resource(column).handler
        table_record = table.get_record(value)
        if table_record is None:
            return None
        return table.get_record_value(table_record, 'title')


    table_actions = []


    def get_table_columns(self, resource, context):
        table_columns = columns[:]
        table_columns.insert(0, ('checkbox', None))
        return table_columns



class Tracker_Search(BaseSearchForm, Tracker_View):

    access = 'is_allowed_to_view'
    title = MSG(u'Search')
    icon = 'search.png'

    # Search Form
    search_template = '/ui/tracker/search.xml'
    search_schema = {
        'search_name': Unicode(),
        'search_title': Unicode(),
        'text': Unicode(),
        'mtime': Integer(),
        'product': Integer(multiple=True),
        'module': Integer(multiple=True),
        'version': Integer(multiple=True),
        'type': Integer(multiple=True),
        'state': Integer(multiple=True),
        'priority': Integer(multiple=True),
        'assigned_to': String(multiple=True)
        }

    context_menus = []

    def get_search_namespace(self, resource, context):
        # Set Style & JS
        context.styles.append('/ui/tracker/tracker.css')
        context.scripts.append('/ui/tracker/tracker.js')

        # Search Form
        get_resource = resource.get_resource
        query = context.query
        search_name = query['search_name']
        if search_name:
            search = get_resource(search_name)
            get_value = search.handler.get_value
            get_values = search.get_values
            search_title = search.get_property('title')
        else:
            get_value = query.get
            get_values = query.get
            search_name = None
            search_title = query['search_title']

        # Build the namespace
        product = get_values('product')
        module = get_values('module')
        version = get_values('version')
        type = get_values('type')
        state = get_values('state')
        priority = get_values('priority')
        assigned_to = get_values('assigned_to')

        # is_admin
        ac = resource.get_access_control()
        pathto_website = resource.get_pathto(resource.get_site_root())

        return  {
           'search_name': search_name,
           'search_title': search_title,
           'text': get_value('text'),
           'mtime': get_value('mtime'),
           'is_admin': ac.is_admin(context.user, resource),
           'manage_assigned': '%s/;browse_users' % pathto_website,
           'products': TrackerList(element='product',
                                   tracker=resource).get_namespace(product),
           'modules': ProductInfoList(element='module',
                                      tracker=resource).get_namespace(module),
           'versions': ProductInfoList(element='version',
                                     tracker=resource).get_namespace(version),
           'types': TrackerList(element='type',
                                tracker=resource).get_namespace(type),
           'states': TrackerList(element='state',
                                 tracker=resource).get_namespace(state),
           'priorities': TrackerList(element='priority',
                                 tracker=resource).get_namespace(priority),
           'assigned_to': UsersList(tracker=resource).get_namespace(
                                                         assigned_to),
           'list_products': resource.get_list_products_namespace()}


    def get_namespace(self, resource, context):
        namespace = BaseSearchForm.get_namespace(self, resource, context)
        namespace['batch'] = None
        namespace['table'] = None
        return namespace



class Tracker_RememberSearch(BaseForm):

    access = 'is_allowed_to_edit'
    schema = merge_dicts(StoredSearchFile.schema,
                         search_name=String,
                         search_title=Unicode(mandatory=True))


    def GET(self, resource, context):
        # Required for when the form fails the automatic checks
        return context.come_back(message=context.message)


    def action(self, resource, context, form):
        search_name = form.get('search_name')
        title = form['search_title']

        # Already a search name ?

        # No
        if search_name is None:
            # Search for a search with the same title
            if isinstance(resource, Issue):
                resource = resource.parent
            searches = resource.search_resources(cls=StoredSearch)
            for search in searches:
                # Found !
                if title == search.get_property('title'):
                    search_name = search.name
                    message = MSG(u'The search has been modified.')
                    break
            else:
                # Not found => so we make a new search resource
                search_name = resource.get_new_id('s')
                search = StoredSearch.make_resource(StoredSearch, resource,
                                                    search_name)
                message = MSG(u'The search has been stored.')
        # Yes
        else:
            search = resource.get_resource(search_name)
            message = MSG(u'The search title has been changed.')

        # Reset the search
        search.handler.load_state_from_string('')

        # Set title
        search.set_property('title', title)

        # Save the value
        for name, type in StoredSearchFile.schema.iteritems():
            value = form[name]
            if value:
                search.set_values(name, value, type)

        # Go
        return context.come_back(message, goto=';view?search_name=%s' %
                                          search_name)



class Tracker_ForgetSearch(BaseForm):

    access = 'is_allowed_to_edit'
    schema = {
        'search_name': String(mandatory=True)}


    def action(self, resource, context, form):
        name = form['search_name']
        resource.del_resource(name)
        # Ok
        message = MSG(u'The search has been removed.')
        return context.come_back(message, goto=';search')



class Tracker_GoToIssue(BaseView):

    access = 'is_allowed_to_view'

    def GET(self, resource, context):
        issue_name = context.get_form_value('issue_name')
        if not issue_name:
            return context.come_back(messages.MSG_NAME_MISSING)

        if not resource.has_resource(issue_name):
            return context.come_back(ERROR(u'Issue not found.'))

        issue = resource.get_resource(issue_name)
        if not isinstance(issue, Issue):
            return context.come_back(ERROR(u'Issue not found.'))

        return context.uri.resolve2('../%s/;edit' % issue_name)



class Tracker_ExportToCSVForm(Tracker_View):

    template = '/ui/tracker/export_to_csv.xml'
    external_form = True

    def get_query_schema(self):
        schema = Tracker_View.get_query_schema(self)
        schema['ids'] = String(multiple=True, default=[])
        return schema


    def get_namespace(self, resource, context):
        namespace = Tracker_View.get_namespace(self, resource, context)
        query = context.query

        # Insert query parameters as hidden input fields
        parameters = []
        schema = Tracker_View.get_query_schema(self)
        for name in schema:
            if name in namespace:
                continue
            value = query[name]
            datatype = schema[name]
            widget = HiddenWidget(name)
            if datatype.multiple is True:
                for value in value:
                    parameters.append(widget.to_html(datatype, value))
            elif value:
                parameters.append(widget.to_html(datatype, value))
        namespace['search_parameters'] = parameters

        return namespace



class Tracker_ExportToCSV(BaseView):

    access = 'is_allowed_to_view'
    title = MSG(u'Export to CSV')
    query_schema = {
        'editor': String(default='excel'),
        'ids': String(multiple=True),
    }


    def GET(self, resource, context):
        # Get search results
        results = resource.get_search_results(context)
        if isinstance(results, Reference):
            return results

        # Selected issues
        issues = results.get_documents()
        selected_issues = context.query['ids']
        if selected_issues:
            issues = [ x for x in issues if x.name in selected_issues ]

        if len(issues) == 0:
            context.message = ERROR(u"No data to export.")
            return

        # Get CSV encoding and separator (OpenOffice or Excel)
        editor = context.query['editor']
        if editor == 'oo':
            separator = ','
            encoding = 'utf-8'
        else:
            separator = ';'
            encoding = 'cp1252'

        # Create the CSV
        csv = CSVFile()
        for issue in issues:
            issue = get_issue_informations(resource, issue)
            row = []
            for name, label in columns:
                value = issue[name]
                if isinstance(value, unicode):
                    value = value.encode(encoding)
                else:
                    value = str(value)
                row.append(value)
            csv.add_row(row)

        # Set response type
        response = context.response
        response.set_header('Content-Type', 'text/comma-separated-values')
        response.set_header('Content-Disposition',
                            'attachment; filename=export.csv')
        return csv.to_str(separator=separator)



class Tracker_ExportToText(Tracker_ExportToCSVForm):

    template = '/ui/tracker/export_to_text.xml'

    def get_query_schema(self):
        schema = Tracker_ExportToCSVForm.get_query_schema(self)
        schema['column_selection'] = String(multiple=True, default=['title'])
        return schema


    def get_namespace(self, resource, context):
        namespace = Tracker_ExportToCSVForm.get_namespace(self, resource,
                                                          context)
        query = context.query

        # Column Selector
        selection = query['column_selection']
        export_columns = columns[2:] + [columns[1]]
        namespace['columns'] = [
            {'name': name, 'title': title, 'checked': name in selection}
            for name, title in export_columns ]

        # Text
        items = self.get_items(resource, context)
        items = self.sort_and_batch(resource, context, items)
        selected_items = query['ids']
        if selected_items:
            items = [ x for x in items if x.name in selected_items ]
        items = [ get_issue_informations(resource, x) for x in items ]
        # Create the text
        lines = []
        for item in items:
            name = item['name']
            line = [u'#%s' % name]
            for x in selection:
                value = item[x]
                if type(value) is unicode:
                    pass
                elif type(value) is str:
                    value = unicode(value, 'utf-8')
                else:
                    value = unicode(value)
                line.append(value)
            line = u'\t'.join(line)
            lines.append(line)
        namespace['text'] = u'\n'.join(lines)

        # Ok
        return namespace



class Tracker_ChangeSeveralBugs(Tracker_View):

    access = 'is_allowed_to_view'
    title = MSG(u'Change Several Issues')
    template = '/ui/tracker/change_bugs.xml'
    schema = {
        'comment': Unicode,
        'ids': String(multiple=True),
        'change_product': Integer,
        'change_module': Integer,
        'change_version': Integer,
        'change_type': Integer,
        'change_priority': Integer,
        'change_assigned_to': String,
        'change_state': Integer,
    }

    external_form = True

    table_actions = [
        Button(name='change_several_bugs', title=MSG(u'Edit issues'))]


    def get_namespace(self, resource, context):
        context.scripts.append('/ui/tracker/tracker.js')
        namespace = Tracker_View.get_namespace(self, resource, context)
        # Edit several bugs at once
        get_resource = resource.get_resource
        namespace['products'] = get_resource('product').get_options()
        namespace['modules'] = get_resource('module').get_options()
        namespace['versions'] = get_resource('version').get_options()
        namespace['priorities'] = get_resource('priority').get_options()
        namespace['types'] = get_resource('type').get_options()
        namespace['states'] = get_resource('state').get_options()
        namespace['users'] = resource.get_members_namespace('')
        namespace['list_products'] = resource.get_list_products_namespace()

        # Ok
        return namespace


    def action(self, resource, context, form):
        # Get search results
        results = resource.get_search_results(context)
        if isinstance(results, Reference):
            return results

        # Selected issues
        issues = results.get_documents()
        selected_issues = form['ids']
        if selected_issues:
            issues = [ x for x in issues if x.name in selected_issues ]

        if len(issues) == 0:
            context.message = ERROR(u"No data to export.")
            return

        # Modify all issues selected
        names = ['product', 'module', 'version', 'type', 'priority', 'state']
        comment = form['comment']
        user = context.user
        username = user and user.name or ''
        users_issues = {}
        for issue in issues:
            issue = resource.get_resource(issue.name)
            # Create a new record
            record = {
                'datetime': datetime.now(),
                'username': username,
                'title': issue.get_value('title'),
                'product': issue.get_value('product'),
                'cc_list': issue.get_value('cc_list'),
                'file': '',
            }
            # Assign-To
            assigned_to = issue.get_value('assigned_to')
            new_assigned_to = form['change_assigned_to']
            if new_assigned_to == 'do-not-change':
                record['assigned_to'] = assigned_to
            else:
                record['assigned_to'] = new_assigned_to
            # Integer Fields
            for name in names:
                new_value = form['change_%s' % name]
                if new_value == -1:
                    record[name] = issue.get_value(name)
                else:
                    record[name] = new_value
            # Comment
            record['comment'] = comment
            modifications = issue.get_diff_with(record, context)
            if modifications:
                title = MSG(u'Modifications:').gettext()
                record['comment'] += u'\n\n%s\n\n%s' % (title, modifications)

            # Save issue
            history = issue.handler.get_handler('.history')
            history.add_record(record)
            # Mail (create a dict with a list of issues for each user)
            info = {'href': context.uri.resolve(issue.name),
                    'name': issue.name,
                    'title': issue.get_title()}
            if new_assigned_to and new_assigned_to != 'do-not-change':
                users_issues.setdefault(new_assigned_to, []).append(info)
            if assigned_to and assigned_to != new_assigned_to:
                users_issues.setdefault(assigned_to, []).append(info)
            # Change
            context.server.change_resource(issue)

        # Send mails
        root = context.root
        if user is None:
            user_title = MSG(u'ANONYMOUS').gettext()
        else:
            user_title = user.get_title()
        template = MSG(u'--- Comment from: $user ---\n\n$comment\n\n$issues')
        tracker_title = resource.get_property('title') or 'Tracker Issue'
        subject = u'[%s]' % tracker_title
        for user_id in users_issues:
            user_issues = [
                u'#%s - %s - %s' % (x['href'], x['name'], x['title'])
                for x in users_issues[user_id]
            ]
            user_issues = '\n'.join(user_issues)
            body = template.gettext(user=user_title, comment=comment,
                                    issues=user_issues)
            to_addr = root.get_user(user_id).get_property('email')
            root.send_email(to_addr, subject, text=body)

        context.message = messages.MSG_CHANGES_SAVED



def get_issue_informations(resource, item):
    """Construct a dict with issue informations.  This dict is used to
    construct a line for a table.
    """
    # Build the namespace
    infos = {
        'name': item.name,
        'id': item.id,
        'title': item.title,
    }

    # Select Tables
    get_resource = resource.get_resource
    tables = ['product', 'module', 'version', 'type', 'state', 'priority']
    for name in tables:
        infos[name] = None
        value = getattr(item, name)
        if value is None:
            continue
        table = get_resource(name).handler
        table_record = table.get_record(value)
        if table_record is None:
            continue
        infos[name] = table.get_record_value(table_record, 'title')

    # Assigned-To
    assigned_to = getattr(item, 'assigned_to')
    infos['assigned_to'] = ''
    if assigned_to:
        users = resource.get_resource('/users')
        if users.has_resource(assigned_to):
            user = users.get_resource(assigned_to)
            infos['assigned_to'] = user.get_title()

    # Modification Time
    mtime_sort = item.mtime
    mtime = datetime.strptime(mtime_sort, '%Y%m%d%H%M%S')
    infos['mtime'] = format_datetime(mtime)
    infos['mtime_sort'] = mtime_sort

    return infos


